#import "osd/osd.hpp"

#define FONT_FILENAME   "etc/shruti_12i.pcx"

int frame_rate = 0;
static int frames_rendered = 0;
static int frame_count = 1;
static volatile int frame_interrupt = FALSE;
static volatile int throttle_counter = 0;

static void fps_timer (void)
{
   frame_interrupt = TRUE;
}
END_OF_STATIC_FUNCTION(fps_timer);

static void throttle_timer (void)
{
   throttle_counter++;
}
END_OF_STATIC_FUNCTION(throttle_timer);

int quit = FALSE;
int reset = FALSE;
static int suspend = FALSE;
static int switched_in = FALSE;
static int switched_out = FALSE;

static void close_button_callback (void)
{
   quit = TRUE;
}

static void switch_in_callback (void)
{
   switched_in = TRUE;
}

static void switch_out_callback (void)
{
   switched_out = TRUE;
}

static void load_sram (void);
static void save_sram (void);
static int confirm (const char *msg);

int __argc;
char **__argv;

int main (int argc, char *argv[])
{
    __argc = argc;
    __argv = argv;

    printf("Genesis XX - Sega Genesis/Mega Drive emulator (v1.0)\n");
    printf("Based upon Genesis Plus\n");
    printf("\n");

    printf("Copyright (C) 1999-2003, Charles MacDonald\n");
    printf("Copyright (C) 2006, Randilyn McDowell\n");
    printf ("Portions Copyright (C) 1998-2001, Karl Stenerud\n");
    printf ("Portions Copyright (C) 1998-2000, Juergen Buchmueller\n");
    printf ("Portions Copyright (C) 2006, Shay Green\n");
    printf ("Portions Copyright (C) 2003, Maxim Stepin\n");
    //printf ("Portions Copyright (C) 2004, The MAME Team\n");
    printf ("Portions Copyright (C) 1998, Gilles Vollant\n");
    printf ("\n");

    printf ("This program is free software; you can redistribute it and/or modify\n");
    printf ("it under the terms of the GNU General Public License as published by\n");
    printf ("the Free Software Foundation; either version 2 of the License, or\n");
    printf ("(at your option) any later version.\n");
    printf ("\n");

    printf ("The Z80 CPU emulator, 68K CPU emulator, and the PSG, YM2612\n");
    printf ("emulation code are taken from the MAME project, and terms of their\n");
    printf ("use are covered under the MAME license.\n");
    printf ("\n");

    printf ("This project uses Musashi, a portable Motorola M680x0 processor\n");
    printf ("emulation engine developed by Karl Stenerud.\n");
    printf ("\n");

    if(argc < 2) {
        printf("Usage %s file.[bin|smd|zip] [-options]\n", argv[0]);
        printf ("\n");
        printf("Type `%s -help' for a list of options.\n", argv[0]);
        exit(1);
    };

    if(stricmp(argv[1], "-help") == 0)
    {
        print_options();
        return (0);
    }

    if(!load_rom(argv[1]))
    {
        printf("File `%s' not found.\n", argv[1]);
        return (0);
    }

    if (init_machine() != 0)
      return (1);

    error_init();
    system_init();

    if(option.sound)
       audio_init (option.sndrate, option.sndvol);

    //load_sram ();
    system_reset();

    throttle_counter=0;

    for(;;)
    {
         int render = FALSE;

        if (quit)
        {
         quit = FALSE;

         if (confirm ("This will exit the emulator."))
            break;
        }

        if (reset)
        {
         reset = FALSE;

         if (confirm ("This will reset the emulator."))
            system_reset();
        }

         if (switched_in)
         {
            switched_in = FALSE;

            suspend = FALSE;

            if (option.sound)
               osd_sound_resume ();

            throttle_counter -= throttle_counter;
         }

         if (switched_out)
         {
            switched_out = FALSE;

            suspend = TRUE;

            if (option.sound)
               osd_sound_pause ();
         }

         if (suspend)
         {
            rest (10);
            continue;
         }

         if (frame_interrupt)
         {
            /* The FPS timer was triggered; sync counters. */

            frame_rate = frames_rendered;
            frames_rendered = 0;

            /* Clear interrupt flag so it doesn't fire again. */
            frame_interrupt = FALSE;
         }

         if (--frame_count > 0)
         {
            /* This frame will be executed, but not drawn. */
            render = FALSE;
         }
         else
         {
            /* This frame will be executed, and drawn. */
            render = TRUE;

            if (option.throttle)
            {
               /* Speed throttling. */
   
               while (throttle_counter == 0)
                  rest (1);
            }

            /* Get all currently pending frames into the frame
               counter. */
            frame_count = throttle_counter;

            /* We use subtract here to avoid losing ticks if the timer
               fires between this and the last statement. */
            throttle_counter -= frame_count;

            /* Enforce frame skip setting if it is not auto. */
            if ((option.skip != -1) &&
                (frame_count > option.skip))
            {
               frame_count = option.skip;
            }
         }

        osd_input_update ();

        if (render)
        {
            system_frame(0);
            osd_video_update ();
            frames_rendered++;
        }
        else
        {
            system_frame(1);
        }

        if(option.sound)
         osd_sound_buffer (snd.buffer[0], snd.buffer[1], snd.buffer_size * 2);

        rest (0);
    }

    system_shutdown();
    //save_sram ();

    error_shutdown();
    trash_machine();
    return (0);
}
END_OF_MAIN()

int load_file(char *filename, char *buf, int size)
{
    FILE *fd = fopen(filename, "rb");
    if(!fd) return (0);
    fread(buf, size, 1, fd);
    fclose(fd);
    return (1);
}

int save_file(char *filename, char *buf, int size)
{
    FILE *fd = NULL;
    if(!(fd = fopen(filename, "wb"))) return (0);
    fwrite(buf, size, 1, fd);
    fclose(fd);
    return (1);
}

static FONT *custom_font = NULL;

int init_machine(void)
{
    char cfg[MAX_FILENAMEZ];
    char title[MAX_FILENAMEZ];

    allegro_init();

    memset (cfg, 0, sizeof (cfg));
    get_executable_name (cfg, (sizeof (cfg) - 1));
    replace_extension (cfg, cfg, "cfg", (sizeof (cfg) - 1));

    do_config(cfg);

    install_keyboard();
    install_mouse ();
    install_joystick(option.joy_driver);

    install_timer();

    if (option.sound)
    {
       if (osd_sound_init () != 0)
         return (1);
    }

    if (osd_video_init() != 0)
      return (2);

    gui_fg_color = makecol (255, 255, 192);
    gui_bg_color = makecol (0, 0, 80);

    custom_font = load_font (FONT_FILENAME, NULL, 0);
    if (custom_font)
      font = custom_font;

   memset (title, 0, sizeof (title));
   strncat (title, "Genesis XX - ", (sizeof (title) - 1));
   strncat (title, get_filename (__argv[1]), (sizeof (title) - 1));
   set_window_title (title);

   set_close_button_callback (close_button_callback);

   set_display_switch_callback (SWITCH_IN, switch_in_callback);
   set_display_switch_callback (SWITCH_OUT, switch_out_callback);

   /* Install timers. */
   install_int_ex (fps_timer, BPS_TO_TIMER(1));
   install_int_ex (throttle_timer, BPS_TO_TIMER(FRAMES_PER_SECOND));

   return (0);
}

void trash_machine(void)
{
   if (custom_font)
   {
      destroy_font (custom_font);
      custom_font = NULL;
   }

    osd_video_shutdown ();

    if(option.sound)
        osd_sound_shutdown();
}


static void load_sram (void)
{
   char filename[MAX_FILENAMEZ];
   PACKFILE *file;

   memset (filename, 0, sizeof (filename));
   replace_extension (filename, get_filename (__argv[1]), "srm", (sizeof (filename) - 1));

   file = pack_fopen (filename, "r");
   if (!file)
      return;

   // TODO: Load SRAM contents here
   pack_fclose (file);

   printf ("Loaded SRAM from %s.\n", filename);
}

static void save_sram (void)
{
   char filename[MAX_FILENAMEZ];
   PACKFILE *file;

   memset (filename, 0, sizeof (filename));
   replace_extension (filename, get_filename (__argv[1]), "srm", (sizeof (filename) - 1));

   file = pack_fopen (filename, "w");
   if (!file)
      return;

   // TODO: Save SRAM contents here
   pack_fclose (file);

   printf ("Saved SRAM to %s.\n", filename);
}

static int confirm (const char *msg)
{
   int result;

   if (option.sound) osd_sound_pause ();

   show_mouse (screen);

   if (alert (msg, NULL, "Are you sure?", "&OK", "&Cancel", 'o', 'c') == 1)
      result = TRUE;
   else
      result = FALSE;

   show_mouse (NULL);

   if (option.sound) osd_sound_resume ();

   throttle_counter -= throttle_counter;

   return (result);
}
