#import "osd/osd.hpp"
#import "osd/hsl.hpp"

/* The Genesis screen which can accomodate up to 1024x512 pixels of data. */
static BITMAP *gen_bmp = NULL;

/* Portion of Genesis screen (the viewport) is copied into this so that it
   can be accessed linearly by the filters and other stuff. */
static BITMAP *in_bmp = NULL;

/* Buffer to be used to hold the output from filters. */
static BITMAP *out_bmp = NULL;

/* Note that 'in_bmp' and 'out_bmp' are destroyed and recreated whenever a
   viewport change occurs.  That code is handled by create_buffers(). */

/* Palette used to store VDP color data. */
static RGB vdp_palette[3][0x200];

/* Screen palette for 256 color modes (option.video_depth == 8). */
static PALETTE palette;
static RGB_MAP map;

/* Stuff for filters. */
static sms_ntsc_setup_t setup;
static sms_ntsc_t* ntsc = NULL;

/* Private functions (defined at bottom). */
static int create_buffers (void);
static void destroy_buffers (void);

static void set_palette_map (RGB *);

#define GET_SCREEN() \
   if(option.vsync) vsync(); \
   acquire_screen ()

#define UNGET_SCREEN()  release_screen ()

static INLINE void draw_scanlines (BITMAP *bmp);
static INLINE void interp (BITMAP *src, BITMAP *dest, int tvmode);

int osd_video_init (void)
{                       
   if (option.video_depth == -1)
      option.video_depth = desktop_color_depth ();
   if (option.video_depth <= 0)
      option.video_depth = 16;

   set_color_depth (option.video_depth);

   /* Try to match the virtual machine's refresh rate. */
   request_refresh_rate (option.video_refresh);

   if (set_gfx_mode (option.video_driver,
                     option.video_width, option.video_height, 0, 0) != 0)
   {
      osd_video_shutdown ();

      allegro_message ("Error setting graphics mode (%dx%d %dbpp).\n"
                       "Allegro says: `%s'\n",
                       option.video_width, option.video_height,
                       option.video_depth,
                       allegro_error);
      return (1);
   }

   /* This is needed because sometimes filters will output video in a color
      format that is not compatible with the screen. */
   set_color_conversion (COLORCONV_TOTAL);

#ifndef ALLEGRO_DOS
   if (is_windowed_mode ())
      set_display_switch_mode (SWITCH_BACKGROUND);
   else
      set_display_switch_mode (SWITCH_BACKAMNESIA);
#endif

   if (option.video_depth == 8)
   {
      if (option.rtpal)
      {
         /* Initialize palette. */
   
         int index;
   
         for (index = 0; index < 0x100; index++)
         {                     
            palette[index].r = (((index >> 5) & 7) << 3);
            palette[index].g = (((index >> 2) & 7) << 3);
            palette[index].b = (((index >> 0) & 3) << 4);
         }
   
         set_palette_map (palette);
      }
      else
      {
         generate_332_palette (palette);
         set_palette_map (palette);
      }
   }

   /* Initialize bitmap data and viewport. */

   memset (&bitmap, 0, sizeof (bitmap));

   bitmap.viewport.w = 256;
   bitmap.viewport.h = 224;
   bitmap.viewport.x = 0x20;
   bitmap.viewport.y = 0x00;

   if (create_buffers () != 0)
   {
      osd_video_shutdown ();

      allegro_message ("Error creating graphics buffers.\n"
                       "Allegro says: `%s'\n",
                       allegro_error);

      return (2);
   }

   /* This code attempts to handle a reversed color order. */
   /* Note that our native color order (renderer & filters) is BGR. */

   /* Setting fixrgb=1 causes the renderer to output RGB instead,
      effectively fixing the problem in a hackish, but efficient way. */

   if (option.fixrgb == -1)
   {
      /* A neat little trick that KittyCat showed me. */

      int depth;
      int red, blue;

      depth = bitmap_color_depth (gen_bmp);

      red = makecol_depth (depth, 255, 0, 0);
      blue = makecol_depth (depth, 0, 0, 255);

      if (red > blue)
         option.fixrgb = 0;
      else
         option.fixrgb = 1;
   }

   return (0);
}

void osd_video_shutdown (void)
{
   destroy_buffers ();

   set_gfx_mode (GFX_TEXT, 0, 0, 0, 0);

   if (ntsc)             
   {
      free (ntsc);
      ntsc = NULL;
   }
}

void osd_video_update (void)
{
    //int swidth = (reg[12] & 1) ? 320 : 256;
    //int sheight = (reg[1] & 8) ? 240 : 224;
    //int twidth = swidth;
    //int theight = sheight;
    int swidth, sheight, twidth, theight;
    int center_x;
    int center_y;

    if(bitmap.viewport.changed)
    {
        bitmap.viewport.changed = 0;

        GET_SCREEN();
        clear_to_color(screen, makecol (0, 0, 0));
        UNGET_SCREEN();

        destroy_buffers ();

        if (create_buffers () != 0)
        {
         allegro_message ("Error creating graphics buffers.\n"
                          "Allegro says: `%s'\n",
                          allegro_error);

         exit (1);
        }
   }

   swidth = bitmap.viewport.w;
   sheight = bitmap.viewport.h;

    switch (option.video_filter)
    {
      case VIDEO_FILTER_FIT:
      {
         twidth = SCREEN_W;
         theight = SCREEN_H;

         break;
      }

      case VIDEO_FILTER_2X:
      case VIDEO_FILTER_INTERP2X:
      case VIDEO_FILTER_TV:
      case VIDEO_FILTER_HQ2X:
      {
         twidth = swidth * 2;
         theight = sheight * 2;
         break;
      }

      case VIDEO_FILTER_NTSC:
      {
         twidth = SMS_NTSC_OUT_WIDTH(swidth);
         theight = sheight * 2;

         break;
      }

      case VIDEO_FILTER_HQ3X:
      {
         twidth = swidth * 3;
         theight = sheight * 3;
         break;
      }

      case VIDEO_FILTER_HQ4X:
      {
         twidth = swidth * 4;
         theight = sheight * 4;
         break;
      }

      default:
      {
         twidth = swidth;
         theight = sheight;

         break;
      }
    }

    center_x = (SCREEN_W - twidth) / 2;
    center_y = (SCREEN_H - theight) / 2;

   if ((option.video_depth == 8) && option.rtpal)
   {
       int set_pal = FALSE;

       if(is_border_dirty)
       {
           uint16 data = *(uint16 *)&cram[(border << 1)];
           RGB *color;
   
           is_border_dirty = 0;
   
           if(reg[12] & 8)
           {
               int j;
               for(j = 0; j < 3; j += 1)
               {
                   color = &vdp_palette[j][data];
                   //set_color((j << 6), color);
                   memcpy (&palette[j<<6], color, sizeof (RGB));
               }
           }
           else
           {
               color = &vdp_palette[1][data];
               //set_color(0x00, color);
               //set_color(0x40, color);
               //set_color(0x80, color);
               memcpy (&palette[0x00], color, sizeof (RGB));
               memcpy (&palette[0x40], color, sizeof (RGB));
               memcpy (&palette[0x80], color, sizeof (RGB));
           }

           set_pal = TRUE;
       }
   
       if(is_color_dirty)
       {
           int i;
           uint16 *p = (uint16 *)&cram[0];
           is_color_dirty = 0;
   
           for(i = 0; i < 64; i += 1)
           {
               if((color_dirty[i]) && ((i & 0x0F) != 0x00))
               {
                   RGB *color;
                   color_dirty[i] = 0;
                   if(reg[12] & 8)
                   {
                       int j;
                       for(j = 0; j < 3; j += 1)
                       {
                           color = &vdp_palette[j][p[i]];
                           //set_color((j << 6) | i, color);
                           memcpy (&palette[j<<6], color, sizeof (RGB));
                       }
                   }
                   else
                   {
                       color = &vdp_palette[1][p[i]];
                       //set_color(0x00 | i, color);
                       //set_color(0x40 | i, color);
                       //set_color(0x80 | i, color);
                        memcpy (&palette[0x00 | i], color, sizeof (RGB));
                        memcpy (&palette[0x40 | i], color, sizeof (RGB));
                        memcpy (&palette[0x80 | i], color, sizeof (RGB));
                   }
               }
           }

           set_pal = TRUE;
       }

       if (set_pal)
         set_palette_map (palette);
    }

    acquire_bitmap (in_bmp);

    // copy viewport
    blit (gen_bmp, in_bmp,
          bitmap.viewport.x, bitmap.viewport.y,
          0, 0,
          bitmap.viewport.w, bitmap.viewport.h);

     if (option.showfps)
     {
         int depth = bitmap_color_depth(in_bmp);
      textprintf_ex (in_bmp, font, 8 + 1, 8 + 1, makecol_depth (depth, 0,0,0), -1, "FPS: %d", frame_rate);
      textprintf_ex (in_bmp, font, 8, 8, makecol_depth (depth, 255,255,255), -1, "FPS: %d", frame_rate);
     }

    release_bitmap (in_bmp);

     switch (option.video_filter)
     {
      case VIDEO_FILTER_NONE:
      {
         if (option.scanlines) draw_scanlines (in_bmp);

         GET_SCREEN();
         blit(in_bmp, screen, 0, 0, center_x, center_y, swidth, sheight);
         UNGET_SCREEN();

         break;
      }

      case VIDEO_FILTER_FIT:
      case VIDEO_FILTER_2X:
      {
         if ((bitmap_color_depth (in_bmp) != bitmap_color_depth (screen)) ||
             option.scanlines)
         {
            acquire_bitmap (out_bmp);
            stretch_blit (in_bmp, out_bmp, 0, 0, swidth, sheight, 0, 0, twidth, theight);
            release_bitmap (out_bmp);
   
            if (option.scanlines) draw_scanlines (out_bmp);
   
            GET_SCREEN();
            blit (out_bmp, screen, 0, 0, center_x, center_y, twidth, theight);
            UNGET_SCREEN();
         }
         else
         {
            GET_SCREEN();
            stretch_blit (in_bmp, screen, 0, 0, swidth, sheight, center_x, center_y, twidth, theight);
            UNGET_SCREEN();
         }

         break;
      }

      case VIDEO_FILTER_INTERP2X:
      case VIDEO_FILTER_TV:
      {
         interp (in_bmp, out_bmp, (option.video_filter == VIDEO_FILTER_TV));

         if (option.scanlines) draw_scanlines (out_bmp);

         GET_SCREEN();
         blit (out_bmp, screen, 0, 0, center_x, center_y, twidth, theight);
         UNGET_SCREEN();

         break;
      }

      case VIDEO_FILTER_NTSC:
      {
         int y;
         unsigned short *pixels_in;
         unsigned short *pixels_out;
         int pitch;

         pixels_in = (unsigned short *)in_bmp->line[0];
         pixels_out = (unsigned short *)out_bmp->line[0];

         pitch = out_bmp->w * 2;

         sms_ntsc_blit (ntsc, pixels_in, in_bmp->w, in_bmp->w, in_bmp->h,
            pixels_out, pitch * 2);

         for (y = 1; y < (out_bmp->h - 1); y += 2)
         {
            unsigned char* io = (unsigned char *)pixels_out + y * pitch;
            int n;

            for ( n = out_bmp->w; n; --n )

               {
                  unsigned prev = *(unsigned short*) (io - pitch);
                  unsigned next = *(unsigned short*) (io + pitch);
                  /* mix 16-bit rgb without losing low bits */
                  unsigned mixed = prev + next + ((prev ^ next) & 0x0821);
                  /* darken by 12% */
                  *(unsigned short*) io = (mixed >> 1) - (mixed >> 4 & 0x18E3);
                  io += 2;
               }
         }

         if (option.scanlines) draw_scanlines (out_bmp);

         GET_SCREEN();
         blit (out_bmp, screen, 0, 0, center_x, center_y, twidth, theight);
         UNGET_SCREEN();

         break;
      }

      case VIDEO_FILTER_HQ2X:
      case VIDEO_FILTER_HQ3X:
      case VIDEO_FILTER_HQ4X:
      {
         int left;

         switch (option.video_filter)
         {
            case VIDEO_FILTER_HQ2X:
            {
               hq2x ((unsigned char *)in_bmp->line[0],
                     (unsigned char *)out_bmp->line[0],
                     in_bmp->w,                        
                     in_bmp->h,
                     (out_bmp->w * 4));

               break;
            }

            case VIDEO_FILTER_HQ3X:
            {
               hq3x ((unsigned char *)in_bmp->line[0],
                     (unsigned char *)out_bmp->line[0],
                     in_bmp->w,                        
                     in_bmp->h,
                     (out_bmp->w * 4));

               break;
            }

            case VIDEO_FILTER_HQ4X:
            {
               hq4x ((unsigned char *)in_bmp->line[0],
                     (unsigned char *)out_bmp->line[0],
                     in_bmp->w,                        
                     in_bmp->h,
                     (out_bmp->w * 4));

               break;
            }
         }

         if (option.scanlines) draw_scanlines (out_bmp);

         GET_SCREEN();
         blit (out_bmp, screen, 0, 0, center_x, center_y, twidth, theight);
         UNGET_SCREEN();

         break;
      }
     }
}

void make_vdp_palette(void)
{
    uint8 lut[3][8] =
    {
        {0x00, 0x04, 0x08, 0x0C, 0x10, 0x14, 0x18, 0x1C},
        {0x00, 0x08, 0x10, 0x18, 0x20, 0x28, 0x30, 0x38},
        {0x20, 0x24, 0x28, 0x2C, 0x30, 0x34, 0x38, 0x3C},
    };

    int j;
    for(j = 0; j < 0x600; j += 1)
    {

        int r = (j >> 6) & 7;
        int g = (j >> 3) & 7;
        int b = (j >> 0) & 7;
        int i = (j >> 9) & 3;

        vdp_palette[i][j & 0x1FF].r = lut[i][r];
        vdp_palette[i][j & 0x1FF].g = lut[i][g];
        vdp_palette[i][j & 0x1FF].b = lut[i][b];
    }
}

/* Private functions. */

int create_buffers (void)
{
   int depth;

   switch (option.video_filter)
   {
      case VIDEO_FILTER_INTERP2X:
      case VIDEO_FILTER_TV:
      case VIDEO_FILTER_NTSC:
      case VIDEO_FILTER_HQ2X:
      case VIDEO_FILTER_HQ3X:
      case VIDEO_FILTER_HQ4X:
      {
         depth = 16;
         break;
      }

      default:
      {
         if (option.video_depth == 8)
         {
            /* We always use a high-color renderer and let Allegro convert
              for us.  A 15-bit mapping table is used for that task. */
            depth = 15;
         }
         else if (option.video_depth == 24)
         {
            /* The renderer does not support 24-bit, so we use 32 and
               let Allegro convert for us. */
            depth = 32;
         }
         else
            depth = option.video_depth;

         break;
      }
   }

   gen_bmp = create_bitmap_ex (depth, 1024, 512);
   if (!gen_bmp)
      return (1);
   clear_bitmap (gen_bmp);

   /* Update bitmap data. */

   bitmap.data = (uint8 *)gen_bmp->line[0];
   bitmap.width = gen_bmp->w;
   bitmap.height = gen_bmp->h;
   bitmap.depth = bitmap_color_depth (gen_bmp);
   bitmap.granularity = ((bitmap.depth + 1) / 8);
   if (bitmap.depth >= 15)
      bitmap.remap = 1;
   bitmap.pitch = (bitmap.width * bitmap.granularity);

   in_bmp = create_bitmap_ex (bitmap_color_depth (gen_bmp),
                              bitmap.viewport.w, bitmap.viewport.h);
   if (!in_bmp)
   {
      destroy_bitmap (gen_bmp);
      return (2);
   }

    make_vdp_palette();

    if (option.rtpal)
    {
       memcpy (palette, black_palette, sizeof (PALETTE));
       palette[0xFE].r =
       palette[0xFE].g =
       palette[0xFE].b = 0x3F;
    }

   switch (option.video_filter)
   {
      case VIDEO_FILTER_FIT:
      case VIDEO_FILTER_2X:
      {
         if ((bitmap_color_depth (in_bmp) != option.video_depth) ||
             option.scanlines)
         {
            int w, h;

            if (option.video_filter == VIDEO_FILTER_2X)
            {
               w = in_bmp->w * 2;
               h = in_bmp->h * 2;
            }
            else
            {
               w = SCREEN_W;
               h = SCREEN_H;
            }

            out_bmp = create_bitmap (w, h);
            if (!out_bmp)
               return (3);
         }

         break;
      }

      case VIDEO_FILTER_INTERP2X:
      case VIDEO_FILTER_TV:
      {
         out_bmp = create_bitmap_ex (16, in_bmp->w*2, in_bmp->h*2);
         if (!out_bmp)
            return (3);

         break;
      }

      case VIDEO_FILTER_NTSC:
      {
         ntsc = (sms_ntsc_t *)malloc (sizeof (sms_ntsc_t));
         if (!ntsc)
            return (3);

         sms_ntsc_init (ntsc, &setup);

         out_bmp = create_bitmap_ex (16, SMS_NTSC_OUT_WIDTH(in_bmp->w), (in_bmp->h * 2));
         if (!out_bmp)
         {
            free (ntsc);
            return (3);
         }

         break;
      }

      case VIDEO_FILTER_HQ2X:
      case VIDEO_FILTER_HQ3X:
      case VIDEO_FILTER_HQ4X:
      {
         int width = 0, height = 0;

         switch (option.video_filter)
         {
            case VIDEO_FILTER_HQ2X:
            {
               width = (in_bmp->w * 2);
               height = (in_bmp->h * 2);

               break;
            }

            case VIDEO_FILTER_HQ3X:
            {
               width = (in_bmp->w * 3);
               height = (in_bmp->h * 3);

               break;
            }

            case VIDEO_FILTER_HQ4X:
            {
               width = (in_bmp->w * 4);
               height = (in_bmp->h * 4);

               break;
            }

            default:
            {
               /* We shouldn't be here. */
               abort ();
               break;
            }
         }

         out_bmp = create_bitmap_ex (32, width, height);
         if (!out_bmp)
            return (3);

         break;
      }

      default:
         break;
   }

   return (0);
}

void destroy_buffers (void)
{
   if (out_bmp)
   {
      destroy_bitmap (out_bmp);
      out_bmp = NULL;
   }

   if (in_bmp)
   {
      destroy_bitmap (in_bmp);
      in_bmp = NULL;
   }

   if (gen_bmp)
   {
      destroy_bitmap (gen_bmp);
      gen_bmp = NULL;
   }
}

void set_palette_map (RGB *pal)
{
   set_palette (pal);

   create_rgb_table (&map, pal, NULL);
   rgb_map = &map;
}

static INLINE void draw_scanlines (BITMAP *bmp)
{
   int depth;
   int line;

   ASSERT(bmp);

   if (!option.scanlines)
      return;

   depth = bitmap_color_depth (bmp);

   set_multiply_blender (0, 0, 0, option.scanlines*255/100);

   drawing_mode (DRAW_MODE_TRANS, NULL, 0, 0);

   acquire_bitmap (bmp);

   for (line = 0; line < bmp->h; line += 2)
      hline (bmp, 0, line, bmp->w, makecol_depth (depth, 0, 0, 0));

   release_bitmap (bmp);

   solid_mode ();
}

#define RED_MASK_16        0x1F
#define GREEN_MASK_16      0x3F
#define BLUE_MASK_16       0x1F

#define RED_SHIFTS_16      3
#define GREEN_SHIFTS_16    2
#define BLUE_SHIFTS_16     3

#define RED_CMASK_16       0xF800
#define GREEN_CMASK_16     0x07E0
#define BLUE_CMASK_16      0X001F

#define RED_CSHIFTS_16     11
#define GREEN_CSHIFTS_16   5
#define BLUE_CSHIFTS_16    0

static INLINE unsigned mix16 (unsigned c1, unsigned c2)
{
   return ((((((c1 & RED_CMASK_16) + (c2 & RED_CMASK_16)) >> 1) & RED_CMASK_16) |
            ((((c1 & GREEN_CMASK_16) + (c2 & GREEN_CMASK_16)) >> 1) & GREEN_CMASK_16) |
            ((((c1 & BLUE_CMASK_16) + (c2 & BLUE_CMASK_16)) >> 1) & BLUE_CMASK_16)));
}

static INLINE unsigned mix16_fast (unsigned c1, unsigned c2)
{
   return (((c1 + c2) >> 1));
}

static INLINE unsigned pack16 (int r, int g, int b)
{
   return ((((r >> RED_SHIFTS_16) << RED_CSHIFTS_16) |
            ((g >> GREEN_SHIFTS_16) << GREEN_CSHIFTS_16) |
            ((b >> BLUE_SHIFTS_16) << BLUE_CSHIFTS_16)));
}

static INLINE void unpack16 (unsigned c, int *r, int *g, int *b)
{
   *r = (((c & RED_CMASK_16) >> RED_CSHIFTS_16) << RED_SHIFTS_16);
   *g = (((c & GREEN_CMASK_16) >> GREEN_CSHIFTS_16) << GREEN_SHIFTS_16);
   *b = (((c & BLUE_CMASK_16) >> BLUE_CSHIFTS_16) << BLUE_SHIFTS_16);
}

static INLINE unsigned add16 (unsigned c1, unsigned c2)
{
   unsigned lsbmask = 0x821;
   unsigned msbmask = 0x8410;
   unsigned lowbits, dest, rbsat, gsat;

   lowbits = (c1 & lsbmask) + (c2 & lsbmask);
   dest = (c1 & ~lsbmask) + (c2 & ~lsbmask) + (lowbits & ~lsbmask);
   dest ^= msbmask * 2;
   rbsat = (((dest & 0x10020) - 0x801) >> 5) & 0xF81F;
   gsat = (((dest & 0x800) - 1) >>  6) & (0x3F <<  5);
   dest = ((dest & ~lsbmask) + (lowbits & lsbmask)) | rbsat | gsat;

   return (dest);
}

static INLINE unsigned sub16 (unsigned c, int amount)
{
   int r, g, b;

   unpack16 (c, &r, &g, &b);

   r -= amount;
   g -= amount;
   b -= amount;
   if (r < 0) r = 0;
   if (g < 0) g = 0;
   if (b < 0) b = 0;

   return (pack16 (r, g, b));
}

static INLINE unsigned pulldown (void)
{
   static uint16 d = rand();
   static int p = 1;
   int d2;

   d = ((d * 50) / p);

   p++;
   if (p > 50)
      p = 1;

  d2 = ((d / 3072) & 0xFF);

  return (pack16 (d2, d2, d2));
}         

static INLINE unsigned mix16_tv (unsigned c1, unsigned c2)
{
   unsigned mixed;

   mixed = mix16 (c1, c2);
   mixed = sub16 (mixed, 30);
   mixed = add16 (mixed, pulldown ());

   return (mixed);
}

static uint16 sat_lut[0xFFFF];
static int sat_lut_built = FALSE;
static void build_saturation_lut (void);

static INLINE void interp (BITMAP *src, BITMAP *dest, int tvmode)
{
   static int phase = 0;
   uint16 *in, *out;
   int w, h, ow, oh;
   int x, y;
                        
   ASSERT(src);
   ASSERT(dest);

   if (!sat_lut_built)
      build_saturation_lut ();

   in = (uint16 *)src->line[0];
   out = (uint16 *)dest->line[0];

   w = src->w;
   h = src->h;
   ow = (src->w * 2);
   oh = (src->h * 2);

   for (y = 0; y < h; y++)
   {
      int yo = ((y * 2) * ow);
      int ym = (y * w);
      unsigned prev = 0;

      for (x = 0; x < w; x++)
      {
         int o = (yo + (x * 2));
         unsigned next, mixed;

         next = in[(ym + x)];
         if (tvmode) next = sat_lut[next];
         mixed = mix16 (prev, next);
         prev = next;

         out[(o + 0)] = mixed;
         out[(o + 1)] = next;
      }
   }

   for (x = 0; x < ow; x++)
   {
      for (y = 1; y < (oh - 1); y += 2)
      {
         unsigned prev, next, mixed;

         prev = out[(((y - 1) * ow) + x)];
         next = out[(((y + 1) * ow) + x)];

         mixed = (tvmode ? mix16_tv (prev, next) : mix16_fast (prev, next));

         out[((y * ow) + x)] = mixed;
      }
   }
}

static void build_saturation_lut (void)
{
   unsigned c;

   for (c = 0; c <= 0xFFFF; c++)
   {
      int r, g, b;
      float h, s, l;
      double diff;

      // TODO: Fix white->black bug.
      //unpack16 (c, &r, &g, &b);
      r = getr16 (c);
      g = getg16 (c);
      b = getb16 (c);

      rgb_to_hsl (r, g, b, &h, &s, &l);

      diff = fabs (h - 0.5) * 2.0;
      diff *= 2;

      s += (s * diff);
      if (s > 1.0f)
         s = 1.0f;

      hsl_to_rgb (h, s, l, &r, &g, &b);

      sat_lut[c] = pack16 (r, g, b);
   }

   sat_lut_built = TRUE;
}
